/*
 timeappend

   Authors:  Hassan Siddiqui - XMM-SOC
         
   Description: Appends a TIME column to a table.
 
   Comments: 
 
  $Id: timeappend.cc,v 1.10 2001/01/09 13:18:39 ulammers Exp $
*/

/* General includes */
#include <algorithm>

/* SAS-specific includes */
#include "param.h"
#include "Dal.h"
#include "CompoundName.h"
#include "OalServer.h"

/* Local includes */
#include "timeappend.h"

/****************************************************
 * class RowData
 *
 *
 * Stores info needed for efficient table processing
 ****************************************************/
class RowData
{
public:
  
  RowData(unsigned row, 
	  unsigned frame, 
	  unsigned ccdNr):
    _row(row),
    _frame(frame),
    _ccdNr(ccdNr)
    {}

  /* Accessors */
  unsigned row()   const { PROC; RETURN _row; }
  unsigned frame() const { PROC; RETURN _frame;}
  unsigned ccdNr() const { PROC; RETURN _ccdNr; }

private:

  unsigned _row;
  unsigned _frame;
  unsigned _ccdNr;
};

/****************************************************
 * class CompRowData
 *
 *
 * Comparison object for sorting RowData
 ****************************************************/

class CompRowData
{
public:
  bool operator()(const RowData & x, const RowData & y)
    {
      PROC;
	  if (x.ccdNr() == y.ccdNr())
	  {
		ALTRETURNV(x.frame() < y.frame());
	  }

      RETURN (x.ccdNr() < y.ccdNr());
    }
};

/****************************************************
 * 
 * Main fn
 *
 ****************************************************/

void timeappend ()
{

  string tablename = stringParameter("table");
  string outset = stringParameter("outset");

  /* Copy dataset to outset */
  DataSet * set = 
	dataSetServer->clone(tablename, outset, Dal::Modify);

  /* Validate compound notation in tablename 
   * and access table component in name */
//    CompoundName cn( tablename, CompoundName::TableName);
//    string tablecomponent = cn.blockName();

  /* Get table name by accessing original table, 
   * using DAL to perform error checking/parsing 
   * of tablename string */
  DataSet * origset = dataSetServer->open(tablename, Dal::Read);
  Table * origtable = origset->table(tablename);
  string tablecomponent = origtable->name();
  dataSetServer->close(origset);

  /* Open new dataset */
  string newtablename = outset + ':' + tablecomponent;
  Table * table = set->table(newtablename);

  Xmm::Instrument instrument;

  if (booleanParameter("withsettings"))
    {
      string instrumentname = stringParameter("instrument");
      errstr << "Setting OAL instrument state to " 
	     << instrumentname
	     << Sas::message(Msg::App, Msg::Sparse);
      instrument = Xmm::toInstrument(instrumentname);
      oalServer.instrument(instrument);
      
      string datamode = stringParameter("datamode");
      errstr << "Setting OAL datamode state to " << datamode
	     << Sas::message(Msg::App, Msg::Sparse);
      oalServer.dataMode(Xmm::toDataMode(datamode));
      
      unsigned expnr = (unsigned) intParameter("expnr");
      
      errstr << "Setting OAL exposure state to " << expnr
	     << Sas::message(Msg::App, Msg::Sparse);
      oalServer.exposureIndex(expnr);

      string node = stringParameter("node");
      
      errstr << "Setting OAL node state to " << node
	     << Sas::message(Msg::App, Msg::Sparse);
      oalServer.node(Xmm::toNode(node));

    }
  else
    {
      
      /* Add dummy attributes CCD_ID QUADRANT 
	 and NODE to avoid oalServer warning */
      string instrumentname = table->attribute("INSTRUME")->asString();
      instrument = Xmm::toInstrument(instrumentname);
      if (instrument == Xmm::EPN)
	{
	  errstr 
	    << "Instrument is EPN. Adding dummy attributes CCDID=0 QUADRANT=0 and CCDNODE=0 to " << outset
	    << Sas::message(Msg::App, Msg::Sparse);
	  *set->addAttribute("CCDID",    "dummy CCD identifier") = 1;
	  *set->addAttribute("QUADRANT", "dummy CCD quadrant") = 0;
	  *set->addAttribute("CCDNODE",  "dummy CCD node") = 0;
	}
      else
	{
	  errstr 
	    << "Instrument either EMOS1, EMOS2 or RGS. Adding dummy attributes CCDID=1 CCDNODE=0 to " << outset
	    << Sas::message(Msg::App, Msg::Sparse);
	  
	  *set->addAttribute("CCDID",    "dummy CCD identifier") = 1;
	  *set->addAttribute("CCDNODE",  "dummy CCD node") = 0;
	}

      /* Set OAL from dataset */
      string outset0 =  outset; 
      errstr << " Setting OAL state using dataset " << outset0
	     << Sas::message(Msg::App, Msg::Sparse);
      oalServer.setState(outset0);
    }
  
  
  /* Get FRAME column */
  Column * frameCol = table->column("FRAME");
  ColumnData * frameColData = frameCol->data();
  int32 * frameData = frameColData->int32Data();

  /* Get CCDID column */
  Column * ccdidCol = table->column("CCDID");
  ColumnData * ccdidColData = ccdidCol->data();
  int8 * ccdidData = ccdidColData->int8Data();

  /* Get QUADRANT column, if necessary */
  Column * quadrantCol;
  ColumnData * quadrantColData = 0;
  int8 * quadrantData = 0;

  bool epn = (instrument == Xmm::EPN);
  if (epn)
    {
      quadrantCol = table->column("QUADRANT");
      quadrantColData = quadrantCol->data();
      quadrantData = quadrantColData->int8Data();
    }

  /* Create TIME column */
  Column * timeCol = 
    table->addColumn(stringParameter("timecolname"), 
		     Column::Real64, 
		     "OBT time", 
		     "secs");
  ColumnData * timeColData = timeCol->data();
  real64 * timeData = timeColData->real64Data();

  /* Create a list of row data */
  vector<RowData> datalist;

  unsigned ccdNr = 0;
  for (unsigned int row = 0; row < timeCol->rows(); ++row)
    {
      if (epn)
		{
		  ccdNr = EPNcam::toChipNr( ccdidData[row], 
					    quadrantData[row]);
		}
      else
		{
		  ccdNr = ccdidData[row];
		}

      datalist.push_back( RowData( row, 
				    frameData[row], 
				    ccdNr)
			   );
    }
  
  /* Sort the list on ccdNr */
  sort(datalist.begin(), datalist.end(), CompRowData());
  
  /* Go through data list, identifying blocks of constant ccdNr */
  unsigned size = datalist.size();
  unsigned index = 0;
  unsigned ccd = 0;
  while (index < size)
    {
      ccd = datalist[index].ccdNr();


      /* Set OAL state according to CCD Nr */
      errstr << " Setting OAL ccd state to  " << ccd
			 << " (based on event list row " << index
			 << ")"
			 << Sas::message(Msg::App, Msg::Sparse);
      oalServer.ccd(ccd);

	  /*** Initialise frame and row vectors */
      vector<int> frameVector;
      vector<int> rowVector;

      while ( (index < size) && (datalist[index].ccdNr() == ccd))
		{
		  frameVector.push_back(datalist[index].frame());
		  rowVector.push_back(datalist[index].row());
		  index++;
		}

      errstr << " index (to next block) " << index
			 << " framesize " << frameVector.size()
			 << Sas::message(Msg::App, Msg::Sparse);
	  
      /*** Process frame vector */

      /* Need to convert frameVector to a C++ array */
      int * frameArray = new int[frameVector.size()];
      for (unsigned i = 0; i < frameVector.size(); ++i)
		{
		  frameArray[i] = frameVector[i];
		}

      const FrameCounterVector * fcVector 
		= oalServer.newFrameCounterVector( frameArray, 
					   frameVector.size());
	  
      ObtVector obtVector = fcVector->toObt();
      vector<double> utcVec(obtVector.toMET());

      
      /* Write info to TIME column */
      for (unsigned i = 0; i < rowVector.size(); ++i)
		{
		  timeData[rowVector[i]] = utcVec[i];
		}
      
    }

  /* Clean up */
  delete timeColData;
  delete frameColData;
  delete ccdidColData;
  if (epn)
    {
      delete quadrantColData;
    }
  dataSetServer->close(set);

}
